from packaging.version import parse, Version
import tox

tox_version = parse(tox.__version__)
assert isinstance(tox_version, Version)

if tox_version.major == 3:
    from tox_docker.tox3.plugin import *
elif tox_version.major == 4:
    from tox_docker.tox4.plugin import *
else:
    raise RuntimeError(f"tox_docker is incomptabile with tox {tox.__version__}")



__all__ = (
    "tox_add_env_config",
    "tox_add_option",
    "tox_after_run_commands",
    "tox_before_run_commands",
)

from typing import List

from tox.config.cli.parser import ToxParser
from tox.config.loader.section import Section
from tox.config.sets import EnvConfigSet
from tox.execute.api import Outcome
from tox.plugin import impl
from tox.session.state import State
from tox.tox_env.api import ToxEnv
from tox.tox_env.errors import Fail

from tox_docker.config import RunningContainers
from tox_docker.plugin import (
    docker_get,
    docker_health_check,
    docker_pull,
    docker_run,
    get_env_vars,
    HealthCheckFailed,
    stop_containers,
)
from tox_docker.tox4.config import DockerConfigSet, parse_container_config
from tox_docker.tox4.log import log


@impl
def tox_add_env_config(env_conf: EnvConfigSet, state: State) -> None:
    def build_docker_config_set(container_name: object) -> DockerConfigSet:
        assert isinstance(container_name, str)
        docker_conf = state.conf.get_section_config(
            section=Section("docker", container_name),
            base=[],
            of_type=DockerConfigSet,
            for_env=None,
        )
        if not docker_conf.loaders:
            raise ValueError(f"Missing [docker:{container_name}] in tox.ini")
        return docker_conf

    env_conf.add_config(
        keys=["docker"],
        of_type=List[DockerConfigSet],
        default=[],
        desc="docker image configs to load",
        factory=build_docker_config_set,  # type: ignore
    )


@impl
def tox_before_run_commands(tox_env: ToxEnv) -> None:
    docker_confs = tox_env.conf.load("docker")

    container_configs = [
        parse_container_config(docker_conf) for docker_conf in docker_confs
    ]

    seen = set()
    for container_config in container_configs:
        if container_config.name in seen:
            raise ValueError(
                f"Container {container_config.name!r} specified more than once"
            )
        seen.add(container_config.name)

    for container_config in container_configs:
        docker_pull(container_config, log)

    config_and_container = []
    running_containers: RunningContainers = {}
    for container_config in container_configs:
        container = docker_run(container_config, running_containers, log)
        config_and_container.append((container_config, container))
        running_containers[container_config.runas_name] = container

    for container_config, container in config_and_container:
        try:
            docker_health_check(container_config, container, log)
        except HealthCheckFailed:
            tox_env.interrupt()
            clean_up_containers(tox_env)
            raise Fail(
                f"{container_config.image!r} (from {container_config.name!r}) failed health check"
            )

        tox_env.conf["set_env"].update(get_env_vars(container_config, container))


@impl
def tox_after_run_commands(
    tox_env: ToxEnv, exit_code: int, outcomes: List[Outcome]
) -> None:
    clean_up_containers(tox_env)


def clean_up_containers(tox_env: ToxEnv) -> None:
    docker_confs = tox_env.conf.load("docker")

    container_configs = [
        parse_container_config(docker_conf) for docker_conf in docker_confs
    ]

    configs_and_containers = []
    for config in container_configs:
        container = docker_get(config)
        if container:
            configs_and_containers.append((config, container))

    stop_containers(configs_and_containers, log)


@impl
def tox_add_option(parser: ToxParser) -> None:
    # command line flag to keep docker containers running
    parser.add_argument(
        "--docker-dont-stop",
        default=[],
        action="append",
        metavar="CONTAINER",
        help=(
            "If specified, tox-docker will not stop CONTAINER after the test run. "
            "Can be specified multiple times."
        ),
    )



from urllib.request import urlopen
import os

import pytest


@pytest.mark.parametrize("instance", ["HEALTHCHECK_BUILTIN", "HEALTHCHECK_CUSTOM"])
def test_the_image_is_healthy(instance: str) -> None:
    host = os.environ[f"{instance}_HOST"]
    port = os.environ[f"{instance}_8000_TCP_PORT"]
    url = f"http://{host}:{port}/healthy"

    response = urlopen(url)
    assert response.getcode() == 200, f"GET {url} => {response.getcode()}"



import os

from tox_docker.plugin import escape_env_var


def test_it_sets_automatic_env_vars() -> None:
    # ksdn117/tcp-udp-test exposes TCP port 1234 and UDP port 5678
    assert "NETWORKING_ONE_HOST" in os.environ
    assert "NETWORKING_ONE_1234_TCP_PORT" in os.environ
    assert "NETWORKING_ONE_5678_UDP_PORT" in os.environ

    # the old names of these variables were dropped in 2.0
    assert "NETWORKING_ONE_1234_TCP" not in os.environ
    assert "NETWORKING_ONE_5678_UDP" not in os.environ


def test_escape_env_var() -> None:
    assert (
        escape_env_var("my.private.registry/cat/image")
        == "MY_PRIVATE_REGISTRY_CAT_IMAGE"
    )



from tox_docker.config import Image


def test_it_parses_name_and_tag() -> None:
    i = Image("nginx:latest")
    assert i.name == "nginx"
    assert i.tag == "latest"


def test_it_defaults_tag_to_None() -> None:
    i = Image("nginx")
    assert i.name == "nginx"
    assert i.tag is None


def test_it_parses_registry_url_into_name() -> None:
    i = Image("docker.io/toxdocker/healthcheck:latest")
    assert i.name == "docker.io/toxdocker/healthcheck"
    assert i.tag == "latest"


def test_it_allows_port_in_registry_url() -> None:
    i = Image("private-registry:5000/namespace/image-name:1.0.0")
    assert i.name == "private-registry:5000/namespace/image-name"
    assert i.tag == "1.0.0"


def test_it_allows_port_in_registry_url_without_tag() -> None:
    i = Image("private-registry:5000/namespace/image-name")
    assert i.name == "private-registry:5000/namespace/image-name"
    assert i.tag is None



from tox.venv import VirtualEnv

from tox_docker.log import LogFunc


def make_logger(venv: VirtualEnv) -> LogFunc:
    def logger(line: str) -> None:
        try:
            # tox 3.7 and later
            action = venv.new_action(line)
        except AttributeError:
            action = venv.session.newaction(venv, line)

        action.setactivity("docker", line)
        with action:
            pass

    return logger



from typing import Collection, Dict, Mapping, Optional
import os
import os.path
import re

from docker.models.containers import Container as DockerContainer
from docker.types import Mount

RunningContainers = Dict[str, DockerContainer]

IMAGE_NAME = re.compile(
    # adapted from https://stackoverflow.com/a/39672069, used under CC-BY-SA
    r"^"
    r"("
    r"(?:(?=[^:\/]{1,253})(?!-)[a-zA-Z0-9-]{1,63}(?<!-)(?:\.(?!-)[a-zA-Z0-9-]{1,63}(?<!-))*(?::[0-9]{1,5})?/)?"
    r"(?![._-])(?:[a-z0-9._-]*)(?<![._-])(?:/(?![._-])[a-z0-9._-]*(?<![._-]))*"
    r")"
    r"(?::((?![.-])[a-zA-Z0-9_.-]{1,128}))?"
    r"$"
)


def runas_name(container_name: str, pid: Optional[int] = None) -> str:
    """
    Generate a name safe for use in parallel scenarios

    This can be either separate invocations of tox (eg on a busy CI server)
    or `tox -p`. The returned name will be stable for a given container name,
    to avoid needing to maintain a mapping.

    """
    pid = pid or os.getpid()
    return f"{container_name}-tox-{pid}"


class Image:
    def __init__(self, config_line: str) -> None:
        match = IMAGE_NAME.match(config_line)
        if not match:
            raise ValueError(f"{config_line!r} is not a valid image name")
        self.name, self.tag = match.groups()

    def __str__(self) -> str:
        if self.tag:
            return f"{self.name}:{self.tag}"
        return self.name

    def __repr__(self) -> str:
        return repr(str(self))


class Port:
    def __init__(self, config_line: str) -> None:
        host_port, _, container_port_proto = config_line.partition(":")
        container_port, _, protocol = container_port_proto.partition("/")

        if protocol.lower() not in ("tcp", "udp"):
            raise ValueError("protocol must be tcp or udp")
        if not host_port.isdigit():
            raise ValueError("host port must be an int")
        if not container_port.isdigit():
            raise ValueError("container port must be an int")

        self.host_port = int(host_port)
        self.container_port_proto = container_port_proto


class Link:
    def __init__(self, config_line: str) -> None:
        target, sep, alias = config_line.partition(":")

        if sep and not alias:
            raise ValueError(f"Link '{target}:' missing alias")

        self.target = runas_name(target)

        # this is what the target will be known as INSIDE the
        # container, so don't substitute the runas_name here
        self.alias = alias or target


class Volume:
    def __init__(self, config_line: str) -> None:
        parts = config_line.split(":")
        if len(parts) != 4:
            raise ValueError(f"Volume {config_line!r} is malformed")
        if parts[0] != "bind":
            raise ValueError(f"Volume {config_line!r} type must be 'bind:'")
        if parts[1] not in ("ro", "rw"):
            raise ValueError(f"Volume {config_line!r} options must be 'ro' or 'rw'")

        volume_type, mode, outside, inside = parts
        if not os.path.isabs(outside):
            raise ValueError(f"Volume source {outside!r} must be an absolute path")
        if not os.path.isabs(inside):
            raise ValueError(f"Mount point {inside!r} must be an absolute path")

        self.docker_mount = Mount(
            source=outside,
            target=inside,
            type=volume_type,
            read_only=bool(mode == "ro"),
        )


class ContainerConfig:
    def __init__(
        self,
        name: str,
        image: Image,
        stop: bool,
        environment: Optional[Mapping[str, str]] = None,
        healthcheck_cmd: Optional[str] = None,
        healthcheck_interval: Optional[float] = None,
        healthcheck_timeout: Optional[float] = None,
        healthcheck_start_period: Optional[float] = None,
        healthcheck_retries: Optional[int] = None,
        ports: Optional[Collection[Port]] = None,
        links: Optional[Collection[Link]] = None,
        volumes: Optional[Collection[Volume]] = None,
    ) -> None:
        self.name = name
        self.runas_name = runas_name(name)
        self.image = image
        self.stop = stop
        self.environment: Mapping[str, str] = environment or {}
        self.ports: Collection[Port] = ports or {}
        self.links: Collection[Link] = links or {}
        self.mounts: Collection[Mount] = [v.docker_mount for v in volumes or ()]

        self.healthcheck_cmd = healthcheck_cmd
        self.healthcheck_interval = (
            int(healthcheck_interval) if healthcheck_interval else None
        )
        self.healthcheck_timeout = (
            int(healthcheck_timeout) if healthcheck_timeout else None
        )
        self.healthcheck_start_period = (
            int(healthcheck_start_period) if healthcheck_start_period else None
        )
        self.healthcheck_retries = healthcheck_retries



from typing import Dict, List

from tox.config.sets import ConfigSet

from tox_docker.config import ContainerConfig, Image, Link, Port, Volume

# nanoseconds in a second; named "SECOND" so that "1.5 * SECOND" makes sense
SECOND = 1000000000


class MissingRequiredSetting(Exception):
    pass


def image_required(image: Image) -> Image:
    if not image.name:
        raise MissingRequiredSetting("image")

    return image


class DockerConfigSet(ConfigSet):
    def register_config(self) -> None:
        self.add_config(
            keys=["image"],
            of_type=Image,
            default=Image(""),
            desc="docker image to run",
            post_process=image_required,
        )
        self.add_config(
            keys=["environment"],
            of_type=Dict[str, str],
            default={},
            desc="environment variables to pass to the docker container",
        )
        self.add_config(
            keys=["ports"],
            of_type=List[Port],
            default=[],
            desc="ports to expose",
        )
        self.add_config(
            keys=["links"],
            of_type=List[Link],
            default=[],
            desc="containers to link",
        )
        self.add_config(
            keys=["volumes"],
            of_type=List[Volume],
            default=[],
            desc="volumes to attach",
        )

        self.add_config(
            keys=["healthcheck_cmd"],
            of_type=str,
            default="",
            desc="docker healthcheck command",
        )
        self.add_config(
            keys=["healthcheck_interval"],
            of_type=float,
            default=0,
            desc="docker healthcheck interval",
            post_process=lambda num: int(num * SECOND),
        )
        self.add_config(
            keys=["healthcheck_timeout"],
            of_type=float,
            default=0,
            desc="docker healthcheck timeout",
            post_process=lambda num: int(num * SECOND),
        )
        self.add_config(
            keys=["healthcheck_start_period"],
            of_type=float,
            default=0,
            desc="docker healthcheck startup grace period",
            post_process=lambda num: int(num * SECOND),
        )
        self.add_config(
            keys=["healthcheck_retries"],
            of_type=int,
            default=0,
            desc="docker healthcheck retry count",
        )


def parse_container_config(docker_config: DockerConfigSet) -> ContainerConfig:
    return ContainerConfig(
        name=docker_config.name,
        image=docker_config["image"],
        stop=docker_config.name not in docker_config._conf.options.docker_dont_stop,
        environment=docker_config["environment"],
        healthcheck_cmd=docker_config["healthcheck_cmd"],
        healthcheck_interval=docker_config["healthcheck_interval"],
        healthcheck_timeout=docker_config["healthcheck_timeout"],
        healthcheck_start_period=docker_config["healthcheck_start_period"],
        healthcheck_retries=docker_config["healthcheck_retries"],
        ports=docker_config["ports"],
        links=docker_config["links"],
        volumes=docker_config["volumes"],
    )



from typing import Container, Dict, Mapping, Optional, Sequence
import re

from tox.config import Config, SectionReader
from tox.venv import VirtualEnv
import py

from tox_docker.config import (
    ContainerConfig,
    Image,
    Link,
    Port,
    RunningContainers,
    Volume,
)

# nanoseconds in a second; named "SECOND" so that "1.5 * SECOND" makes sense
SECOND = 1000000000


EnvRunningContainers = Dict[VirtualEnv, RunningContainers]


def getfloat(reader: SectionReader, key: str) -> Optional[float]:
    val = reader.getstring(key)
    if val is None:
        return None

    try:
        return float(val)
    except ValueError:
        msg = f"{val!r} is not a number (for {key} in [{reader.section_name}])"
        raise ValueError(msg)


def gettime(reader: SectionReader, key: str) -> Optional[int]:
    raw = getfloat(reader, key)
    if raw is None:
        return None

    return int(raw * SECOND)


def getint(reader: SectionReader, key: str) -> Optional[int]:
    raw = getfloat(reader, key)
    if raw is None:
        return None

    val = int(raw)
    if val != raw:
        msg = f"{val!r} is not an int (for {key} in [{reader.section_name}])"
        raise ValueError(msg)
    return val


def getenvdict(reader: SectionReader, key: str) -> Mapping[str, str]:
    environment = {}
    for value in reader.getlist(key):
        envvar, _, value = value.partition("=")
        environment[envvar] = value
    return environment


def discover_container_configs(config: Config) -> Sequence[str]:
    """
    Read the tox.ini, and return a list of docker container configs.

    """

    inipath = str(config.toxinipath)
    iniparser = py.iniconfig.IniConfig(inipath)

    container_names = set()
    for section in iniparser.sections:
        if not section.startswith("docker:"):
            continue

        _, _, container_name = section.partition(":")
        if not re.match(r"^[a-zA-Z][-_.a-zA-Z0-9]+$", container_name):
            raise ValueError(f"{container_name!r} is not a valid container name")

        # populated in the next loop
        container_names.add(container_name)

    return list(container_names)


def parse_container_config(
    config: Config, container_name: str, all_container_names: Container[str]
) -> ContainerConfig:
    inipath = str(config.toxinipath)
    iniparser = py.iniconfig.IniConfig(inipath)

    reader = SectionReader(f"docker:{container_name}", iniparser)
    reader.addsubstitutions(
        distdir=config.distdir,
        homedir=config.homedir,
        toxinidir=config.toxinidir,
        toxworkdir=config.toxworkdir,
    )

    kwargs = {
        "name": container_name,
        "image": Image(reader.getstring("image")),
        "stop": container_name not in config.option.docker_dont_stop,
    }

    environment = None
    if reader.getstring("environment"):
        environment = getenvdict(reader, "environment")

    hc_cmd = hc_interval = hc_timeout = hc_start_period = hc_retries = None
    if reader.getstring("healthcheck_cmd"):
        hc_cmd = reader.getstring("healthcheck_cmd")
    if reader.getstring("healthcheck_interval"):
        hc_interval = gettime(reader, "healthcheck_interval")
    if reader.getstring("healthcheck_timeout"):
        hc_timeout = gettime(reader, "healthcheck_timeout")
    if reader.getstring("healthcheck_start_period"):
        hc_start_period = gettime(reader, "healthcheck_start_period")
    if reader.getstring("healthcheck_retries"):
        hc_retries = getint(reader, "healthcheck_retries")

    ports = None
    if reader.getstring("ports"):
        ports = [Port(line) for line in reader.getlist("ports")]

    links = None
    if reader.getstring("links"):
        links = [Link(line) for line in reader.getlist("links")]

    volumes = None
    if reader.getstring("volumes"):
        volumes = [Volume(line) for line in reader.getlist("volumes")]

    return ContainerConfig(
        name=container_name,
        image=Image(reader.getstring("image")),
        stop=container_name not in config.option.docker_dont_stop,
        environment=environment,
        healthcheck_cmd=hc_cmd,
        healthcheck_interval=hc_interval,
        healthcheck_timeout=hc_timeout,
        healthcheck_start_period=hc_start_period,
        healthcheck_retries=hc_retries,
        ports=ports,
        links=links,
        volumes=volumes,
    )



import os


def test_the_image_is_healthy() -> None:
    # the healthcheck creates a file "healthy" in the volume from within
    # the container; this test proves it's visible outside the container,
    # and thus the bind mount worked as expected
    volume = os.environ["VOLUME_DIR"]
    assert "healthy" in os.listdir(volume)



__all__ = (
    "tox_configure",
    "tox_runtest_pre",
    "tox_runtest_post",
    "tox_addoption",
)

from typing import Dict
import os

from tox import hookimpl
from tox.action import Action
from tox.config import Config, Parser
from tox.venv import VirtualEnv

from tox_docker.config import ContainerConfig, RunningContainers
from tox_docker.plugin import (
    docker_health_check,
    docker_pull,
    docker_run,
    get_env_vars,
    HealthCheckFailed,
    stop_containers,
)
from tox_docker.tox3.config import (
    discover_container_configs,
    EnvRunningContainers,
    parse_container_config,
)
from tox_docker.tox3.log import make_logger

CONTAINER_CONFIGS: Dict[str, ContainerConfig] = {}
ENV_CONTAINERS: EnvRunningContainers = {}


def _newaction(venv: VirtualEnv, message: str) -> Action:
    try:
        # tox 3.7 and later
        return venv.new_action(message)
    except AttributeError:
        return venv.session.newaction(venv, message)


def _config_name(running_name: str) -> str:
    # inverse of tox_docker.config.runas_name()
    suffix = f"-tox-{os.getpid()}"
    assert running_name.endswith(suffix)
    return running_name[: -len(suffix)]


@hookimpl
def tox_configure(config: Config) -> None:
    container_config_names = discover_container_configs(config)

    # validate command line options
    for container_name in config.option.docker_dont_stop:
        if container_name not in container_config_names:
            raise ValueError(
                f"Container {container_name!r} not found (from --docker-dont-stop)"
            )

    for container_name in container_config_names:
        CONTAINER_CONFIGS[container_name] = parse_container_config(
            config, container_name, container_config_names
        )


@hookimpl
def tox_runtest_pre(venv: VirtualEnv) -> None:
    envconfig = venv.envconfig
    container_names = envconfig.docker

    log = make_logger(venv)

    env_container_configs = []

    seen = set()
    for container_name in container_names:
        if container_name not in CONTAINER_CONFIGS:
            raise ValueError(f"Missing [docker:{container_name}] in tox.ini")
        if container_name in seen:
            raise ValueError(f"Container {container_name!r} specified more than once")
        seen.add(container_name)
        env_container_configs.append(CONTAINER_CONFIGS[container_name])

    for container_config in env_container_configs:
        docker_pull(container_config, log)

    ENV_CONTAINERS.setdefault(venv, {})
    running_containers = ENV_CONTAINERS[venv]

    for container_config in env_container_configs:
        container = docker_run(container_config, running_containers, log)
        running_containers[container_config.runas_name] = container

    for running_name, container in running_containers.items():
        container_name = _config_name(running_name)
        container_config = CONTAINER_CONFIGS[container_name]
        try:
            docker_health_check(container_config, container, log)
        except HealthCheckFailed:
            msg = f"{container_config.image!r} (from {container_name!r}) failed health check"
            venv.status = msg
            tox_runtest_post(venv)
            raise

    for running_name, container in running_containers.items():
        container_name = _config_name(running_name)
        container_config = CONTAINER_CONFIGS[container_name]
        for key, val in get_env_vars(container_config, container).items():
            venv.envconfig.setenv[key] = val


@hookimpl
def tox_runtest_post(venv: VirtualEnv) -> None:
    env_containers: RunningContainers = ENV_CONTAINERS.get(venv, {})
    containers_and_configs = [
        (CONTAINER_CONFIGS[_config_name(name)], container)
        for name, container in env_containers.items()
    ]
    log = make_logger(venv)
    stop_containers(containers_and_configs, log)


@hookimpl
def tox_addoption(parser: Parser) -> None:
    # necessary to allow the docker= directive in testenv sections
    parser.add_testenv_attribute(
        name="docker",
        type="line-list",
        help="Name of docker images, including tag, to start before the test run",
        default=[],
    )

    # command line flag to keep docker containers running
    parser.add_argument(
        "--docker-dont-stop",
        default=[],
        action="append",
        metavar="CONTAINER",
        help=(
            "If specified, tox-docker will not stop CONTAINER after the test run. "
            "Can be specified multiple times."
        ),
    )



from contextlib import contextmanager
from typing import Iterator
from unittest.mock import patch
import sys

import pytest

from tox_docker.plugin import get_gateway_ip


@contextmanager
def sys_platform_as(value: str) -> Iterator[None]:
    old_value = sys.platform
    sys.platform = value
    try:
        yield
    finally:
        sys.platform = old_value


class NotARealContainer(object):
    attrs = {"NetworkSettings": {"Gateway": "1.2.3.4"}}


def test_gateway_ip_is_read_from_container_attrs_on_linux() -> None:
    container = NotARealContainer()
    with sys_platform_as("linux2"):
        assert get_gateway_ip(container) == "1.2.3.4"


def test_gateway_ip_is_zero_zero_zero_zero_on_macos() -> None:
    container = NotARealContainer()
    with sys_platform_as("darwin"):
        assert get_gateway_ip(container) == "0.0.0.0"


@pytest.mark.parametrize("sys_platform", ["linux2", "darwin"])
def test_gateway_ip_from_env_var_overrides_attrs_on_linux(sys_platform: str) -> None:
    container = NotARealContainer()
    with sys_platform_as(sys_platform):
        with patch.dict("os.environ", {"TOX_DOCKER_GATEWAY": "192.168.1.1"}):
            assert get_gateway_ip(container) == "192.168.1.1"


@pytest.mark.parametrize("sys_platform", ["linux2", "darwin"])
def test_gateway_ip_is_looked_up_from_hostname(sys_platform: str) -> None:
    container = NotARealContainer()
    with sys_platform_as(sys_platform):
        with patch.dict("os.environ", {"TOX_DOCKER_GATEWAY": "localhost"}):
            assert get_gateway_ip(container) in {"127.0.0.1", "::1"}



from setuptools import find_packages, setup

setup(
    name="tox-docker",
    description="Launch a docker instance around test runs",
    long_description=open("README.rst").read(),
    url="https://github.com/tox-dev/tox-docker",
    maintainer="Dan Crosta",
    maintainer_email="dcrosta@late.am",
    install_requires=[
        "docker>=2.3.0,<6.0",
        "packaging",
        "tox>=3.0.0,<5.0",
    ],
    packages=find_packages(),
    entry_points={"tox": ["docker = tox_docker"]},
    vcversioner={"version_module_paths": ["_version.py"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Testing",
    ],
)



from typing import Dict, Iterable, List, Mapping, Optional, Tuple, Union
import os
import socket
import sys
import time

from docker.errors import ImageNotFound, NotFound
from docker.models.containers import Container
import docker as docker_module

from tox_docker.config import ContainerConfig, RunningContainers
from tox_docker.log import LogFunc


class HealthCheckFailed(Exception):
    pass


def get_gateway_ip(container: Container) -> str:
    gateway = os.getenv("TOX_DOCKER_GATEWAY")
    if gateway:
        ip = socket.gethostbyname(gateway)
    elif sys.platform == "darwin":
        # https://docs.docker.com/docker-for-mac/networking/#use-cases-and-workarounds:
        # there is no bridge network available in Docker for Mac, and exposed ports are
        # made available on localhost (but 0.0.0.0 works just as well)
        ip = "0.0.0.0"
    else:
        ip = container.attrs["NetworkSettings"]["Gateway"] or "0.0.0.0"
    return ip


def escape_env_var(varname: str) -> str:
    """
    Convert a string to a form suitable for use as an environment variable.

    The result will be all uppercase, and will have all invalid characters
    replaced by an underscore.

    The result will match the following regex: [a-zA-Z_][a-zA-Z0-9_]*

    Example:
        "my.private.registry/cat/image" will become
        "MY_PRIVATE_REGISTRY_CAT_IMAGE"
    """
    varletters = list(varname.upper())
    if not varletters[0].isalpha():
        varletters[0] = "_"
    for i, c in enumerate(varletters):
        if not c.isalnum() and c != "_":
            varletters[i] = "_"
    return "".join(varletters)


def docker_pull(container_config: ContainerConfig, log: LogFunc) -> None:
    docker = docker_module.from_env(version="auto")

    try:
        docker.images.get(container_config.image.name)
    except ImageNotFound:
        log(f"pull {container_config.image!r} (from {container_config.name!r})")
        docker.images.pull(container_config.image.name, tag=container_config.image.tag)


def docker_run(
    container_config: ContainerConfig,
    running_containers: RunningContainers,
    log: LogFunc,
) -> Container:
    docker = docker_module.from_env(version="auto")

    healthcheck: Dict[str, Union[List[str], int]] = {}
    if container_config.healthcheck_cmd:
        healthcheck["test"] = ["CMD-SHELL", container_config.healthcheck_cmd]
    if container_config.healthcheck_interval:
        healthcheck["interval"] = container_config.healthcheck_interval
    if container_config.healthcheck_timeout:
        healthcheck["timeout"] = container_config.healthcheck_timeout
    if container_config.healthcheck_start_period:
        healthcheck["start_period"] = container_config.healthcheck_start_period
    if container_config.healthcheck_retries:
        healthcheck["retries"] = container_config.healthcheck_retries

    ports = {p.container_port_proto: p.host_port for p in container_config.ports}

    links = {}
    for link in container_config.links:
        if link.target not in running_containers:
            raise ValueError(
                f"Container {link.target!r} not running; it must come before {container_config.name!r} in the docker= list"
            )
        other_container = running_containers[link.target]
        links[other_container.id] = link.alias

    for mount in container_config.mounts:
        source = mount["Source"]
        if not os.path.exists(source):
            raise ValueError(f"Volume source {source!r} does not exist")

    log(f"run {container_config.image!r} (from {container_config.name!r})")
    container = docker.containers.run(
        str(container_config.image),
        name=container_config.runas_name,
        detach=True,
        environment=container_config.environment,
        healthcheck=healthcheck or None,
        links=links,
        ports=ports,
        publish_all_ports=len(ports) == 0,
        mounts=container_config.mounts,
    )
    container.reload()  # TODO: why do we need this?
    return container


def docker_health_check(
    container_config: ContainerConfig, container: Container, log: LogFunc
) -> None:
    docker = docker_module.from_env(version="auto")

    if "Health" in container.attrs["State"]:
        log(f"health check {container_config.image!r} (from {container_config.name!r})")
        while True:
            container.reload()
            health = container.attrs["State"]["Health"]["Status"]
            if health == "healthy":
                break
            elif health == "starting":
                time.sleep(0.1)
            elif health == "unhealthy":
                # the health check failed after its own timeout
                msg = f"{container_config.image!r} (from {container_config.name!r}) failed health check"
                raise HealthCheckFailed(msg)


def docker_stop(
    container_config: ContainerConfig, container: Container, log: LogFunc
) -> None:
    if container_config.stop:
        log(f"remove '{container.short_id}' (from {container_config.name!r})")
        container.remove(v=True, force=True)
    else:
        log(f"leave '{container.short_id}' (from {container_config.name!r}) running")


def docker_get(container_config: ContainerConfig) -> Optional[Container]:
    docker = docker_module.from_env(version="auto")
    try:
        return docker.containers.get(container_config.runas_name)
    except NotFound:
        return None


def stop_containers(
    containers: Iterable[Tuple[ContainerConfig, Container]], log: LogFunc
) -> None:
    for container_config, container in containers:
        docker_stop(container_config, container, log)


def get_env_vars(
    container_config: ContainerConfig, container: Container
) -> Mapping[str, str]:
    env = {}
    gateway_ip = get_gateway_ip(container)
    for containerport, hostports in container.attrs["NetworkSettings"]["Ports"].items():
        if hostports is None:
            # The port is exposed by the container, but not published.
            continue

        for spec in hostports:
            if spec["HostIp"] == "0.0.0.0":
                hostport = spec["HostPort"]
                break
        else:
            continue

        env[escape_env_var(f"{container_config.name}_HOST")] = gateway_ip
        env[escape_env_var(f"{container_config.name}_{containerport}_PORT")] = hostport

    return env



from urllib.request import urlopen
import os


def test_exposed_ports_are_accessible_to_test_runs() -> None:
    host = os.environ["HEALTHCHECK_BUILTIN_HOST"]
    port = os.environ["HEALTHCHECK_BUILTIN_8000_TCP_PORT"]

    response = urlopen(f"http://{host}:{port}/")
    assert response.getcode() == 200
    assert b"Directory listing for /" in response.read()


def test_it_exposes_only_specified_port() -> None:
    # this container remaps 1234/tcp to 2345, and hides 5678/udp
    assert os.environ["NETWORKING_TWO_1234_TCP_PORT"] == "2345"
    assert "NETWORKING_TWO_5678_UDP_PORT" not in os.environ


def test_docker_picks_a_port_when_you_map_to_zero() -> None:
    # tox.ini has `ports = 0:1234/tcp 0:5678/udp`
    assert os.environ["NETWORKING_ONE_1234_TCP_PORT"] != "0"
    assert os.environ["NETWORKING_ONE_5678_UDP_PORT"] != "0"
